#include "myglwidget.h"
#include "controller.h"

//QGLShaderProgram *MyGLWidget::phongShader;

MyGLWidget::MyGLWidget(int index, Controller* c, bool isPerspective, QQuaternion initialRotation, QWidget *parent)
    : QGLWidget(QGLFormat(QGL::SampleBuffers), parent)
{
    this->index = index;
    MyGLWidget::modeCamera = true;
    tesselation = 1;
    qubeRotation = QQuaternion();
    cameraRotation = initialRotation;
    cameraStartRotation = initialRotation;
    cubeTranslation = QVector3D(0,0,0);
    camRotCenter = QVector3D(0,0,0);
    cameraZoom = cameraZoomDefault;
    clickStartPosition = QPoint(0,0);
    rightButtonPressed = false;
    controller = c;
    this->isPerspective = isPerspective;
    this->isFocused = isPerspective; //start with focus on perspective view
}

MyGLWidget::~MyGLWidget()
{
    /*http://doc.qt.io/qt-4.8/qt-opengl-hellogl-example.html :
    * We use a destructor to ensure that any OpenGL-specific
    * data structures are deleted when the widget is no longer
    * needed (although in this case nothing needs cleaning up).
    */
    //In helloGL example: nothing needs cleaning up.
}

//0 = perspective, 1 = front, 2 = left, 3 = top. Hardcoded on construction.
int MyGLWidget::getIndex(){
    return index;
}

/*
 * From Example
 * We provide size hint functions to ensure that
 * the widgetis shown at a reasonable size:
 */
QSize MyGLWidget::minimumSizeHint() const
{
    return QSize(50, 50);
}

/*
 * From Example
 * We provide size hint functions to ensure that
 * the widgetis shown at a reasonable size:
 */
QSize MyGLWidget::sizeHint() const
{
    return QSize(600, 400);
}


void MyGLWidget::initializeGL()
{
    glClearColor(0.0, 0.0, 0.0, 0.0);
    glEnable(GL_DEPTH_TEST);
    qDebug("InitGPL");
//    if (!phongShader){
        qDebug("Init shader stuff");
        initShaderStuff();
//    }else {
//        qDebug() << " phong not null: " << phongShader;
//    }


//    glEnable(GL_CULL_FACE);//backface culling.
//    glCullFace(GL_BACK);//backface culling

//    glEnable(GL_COLOR_MATERIAL); //Habe ich für glColor3f gebraucht. Wieder raus für Material. //In order to use both color and lighting you must activate openGL color materials

    //4.1.1
    /*
     * The initial state of the OpenGL state machine is defined in
initializeGL(). Furthermore, you can set all OpenGL parameters
that will not change during program execution here as well, for
example, the background color.
     */

    glEnable(GL_LIGHTING);//enable lighting.
    glEnable(GL_LIGHT0);//enable an OpenGL light
 /*
  * https://msdn.microsoft.com/en-us/library/windows/desktop/dd373578%28v=vs.85%29.aspx
  * The glLightfv function returns light source parameter values.
  * GL_POSITION:
  * The params parameter contains four floating-point values
  * that specify the position of the light in homogeneous object
  * coordinates. Both integer and floating-point values are
  * mapped directly. Neither integer nor floating-point values
  * are clamped.
 */
    static GLfloat lightPosition[4] = {0.5, 0.0, 2.0, 1.0};
    glLightfv(GL_LIGHT0,GL_POSITION,lightPosition);//set Position of light source.
}

void MyGLWidget::initShaderStuff(){
    //Phong Shader:
    QGLShader* vert = new QGLShader(QGLShader::Vertex);
    qDebug() << "new vert shader";
    QGLShader* frag = new QGLShader(QGLShader::Fragment);
    qDebug() << "new frag shader";

    bool e = QFile::exists(":/phong.vert");

    qDebug() << "vert exists" << e;

    bool successVertex = vert->compileSourceFile(":/phong.vert");
    if (!successVertex)qDebug() << "Vertex compilation failed.";
    else qDebug() << "Vertex compiled.";
    bool successFragment = frag->compileSourceFile(":/phong.frag");
    if (!successFragment) qDebug() << "Frag failed";
    else qDebug() << "frag success";

    phongShader = new QGLShaderProgram(this);
    phongShader->addShader(vert);
    phongShader->addShader(frag);
    phongShader->link();


    shadeFlatSlot();//set up flat shading
    qDebug() << "FlatSlot setup.";
}

void MyGLWidget::paintGL(){
    qDebug("paint gl");
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glLoadIdentity();

    //cam rotation
    QMatrix4x4 camRot = QMatrix4x4();
    camRot.rotate(cameraRotation);
    QMatrix4x4 zoomCameraMatrix = QMatrix4x4(1,0,0,0,0,1,0,0,0,0,1,cameraZoom,0,0,0,1);

    //So scheint das zumindest für die Kamera zu funktionieren:
    //1. ins Rotationszentrum verschieben - Wenn Rotationszentrum geändert wird, schon rotation berücksichtigen
    //2. Rotieren
    //3. Zoomen
    glMultMatrixf(zoomCameraMatrix.data());
    glMultMatrixf(camRot.data());
    glTranslatef(camRotCenter.x(),camRotCenter.y(),camRotCenter.z());


    //for cube rotation
    QMatrix4x4 m = QMatrix4x4();
    m.rotate(qubeRotation);

    //for cube translation
    QMatrix4x4 translateRot1 = QMatrix4x4(1,0,0,cubeTranslation.x(),0,1,0,cubeTranslation.y(),0,0,1,cubeTranslation.z(),0,0,0,1);

    //stack matrices
    glMultMatrixf(translateRot1.data());//Punkte zurück schieben damit rotation um qube zentrum ist
    glMultMatrixf(m.data());



//tesselation
    double qubeWidth = 1.0;
    double dist = qubeWidth/(double)tesselation;



    //4.1.1 Unit Qube + Tesselation + Diffuse Material
    {

    GLfloat specularColor[] = {1,1,1};
    GLfloat shininess[] = {128};//specular exponent
    glMaterialfv(GL_FRONT,GL_SPECULAR,specularColor);
    glMaterialfv(GL_FRONT,GL_SHININESS,shininess);

    //vorne // z-Achse
//    glColor3f(0,0,1);//blau //alt und doof
    GLfloat blue[] = {0,0,1};
    glMaterialfv(GL_FRONT,GL_DIFFUSE,blue);
    glNormal3f(0,0,1);
    glBegin(GL_QUADS);
    for (int i = 0; i < tesselation; i++){
        for (int j = 0; j < tesselation; j++){
            glVertex3f(0.5-dist*i,0.5-dist*j,0.5);
            glVertex3f(0.5-dist*(i+1),0.5-dist*j,0.5);
            glVertex3f(0.5-dist*(i+1),0.5-dist*(j+1),0.5);
            glVertex3f(0.5-dist*i,0.5-dist*(j+1),0.5);
        }
    }
    glEnd();

    //oben // y-Achse
    GLfloat green[] = {0,1,0};
    glMaterialfv(GL_FRONT,GL_DIFFUSE,green);
    glNormal3f(0,1,0);
    glBegin(GL_QUADS);
    for (int i = 0; i < tesselation; i++){
        for (int j = 0; j < tesselation; j++){
            glVertex3f(0.5-dist*(i+1), 0.5, 0.5-dist*(j+1));
            glVertex3f(0.5-dist*(i+1), 0.5, 0.5-dist*j);
            glVertex3f(0.5-dist*i, 0.5, 0.5-dist*j);
            glVertex3f(0.5-dist*i, 0.5, 0.5-dist*(j+1));
        }
    }
    glEnd();

    //rechts //x-Achse
    GLfloat red[] = {1,0,0};
    glMaterialfv(GL_FRONT,GL_DIFFUSE,red);
    glNormal3f(1,0,0);
    glBegin(GL_QUADS);
    for (int i = 0; i < tesselation; i++){
        for (int j = 0; j < tesselation; j++){
            glVertex3f(0.5, 0.5-dist*i, 0.5-dist*j);
            glVertex3f(0.5, 0.5-dist*(i+1), 0.5-dist*j);
            glVertex3f(0.5, 0.5-dist*(i+1), 0.5-dist*(j+1));
            glVertex3f(0.5, 0.5-dist*i, 0.5-dist*(j+1));
        }
    }
    glEnd();

    //hinten // z-Achse
    GLfloat yellow[] = {1,0.9f,0};
    glMaterialfv(GL_FRONT,GL_DIFFUSE,yellow);
    glNormal3f(0,0,-1);
    glBegin(GL_QUADS);
    for (int i = 0; i < tesselation; i++){
        for (int j = 0; j < tesselation; j++){
            glVertex3f(0.5-dist*i, 0.5-dist*(j+1), -0.5);//4
            glVertex3f(0.5-dist*(i+1), 0.5-dist*(j+1), -0.5);//3
            glVertex3f(0.5-dist*(i+1), 0.5-dist*j, -0.5);//2
            glVertex3f(0.5-dist*i, 0.5-dist*j, -0.5);//1
        }
    }
    glEnd();

    //links
    GLfloat cyan[] = {0,1,1};
    glMaterialfv(GL_FRONT,GL_DIFFUSE,cyan);
    glNormal3f(-1,0,0);
    glBegin(GL_QUADS);
    for (int i = 0; i < tesselation; i++){
        for (int j = 0; j < tesselation; j++){
            glVertex3f(-0.5, 0.5-dist*i, 0.5-dist*(j+1));
            glVertex3f(-0.5, 0.5-dist*(i+1), 0.5-dist*(j+1));
            glVertex3f(-0.5, 0.5-dist*(i+1), 0.5-dist*j);
            glVertex3f(-0.5, 0.5-dist*i, 0.5-dist*j);
        }
    }
    glEnd();

    //unten //-y
    GLfloat magenta[] = {1,0,1};
    glMaterialfv(GL_FRONT,GL_DIFFUSE,magenta);
    glNormal3f(0,-1,0);
    glBegin(GL_QUADS);
    for (int i = 0; i < tesselation; i++){
        for (int j = 0; j < tesselation; j++){
            glVertex3f(0.5-dist*i, -0.5, 0.5-dist*(j+1));
            glVertex3f(0.5-dist*i, -0.5, 0.5-dist*j);
            glVertex3f(0.5-dist*(i+1), -0.5, 0.5-dist*j);
            glVertex3f(0.5-dist*(i+1), -0.5, 0.5-dist*(j+1));
        }
    }
    glEnd();
    }

    //has to be down here or will be overpainted by above code
    if (isFocused) highlightViewport();

}

void MyGLWidget::setFocused(bool isFocused){
    this->isFocused = isFocused;
    updateGL();
}

void MyGLWidget::highlightViewport(){
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(-1,1,-1,1, -1, 1);//to draw yellow rectangle in screen coordinates

    GLfloat specularColor[] = {1,1,1};
    GLfloat shininess[] = {128};//specular exponent
    glMaterialfv(GL_FRONT,GL_SPECULAR,specularColor);
    glMaterialfv(GL_FRONT,GL_SHININESS,shininess);

    GLfloat yellow[] = {1,1,0};
    glMaterialfv(GL_FRONT,GL_DIFFUSE,yellow);
    glNormal3f(0,0,1);

    glDisable(GL_DEPTH_TEST);
    glLineWidth(15.0f);
    glBegin(GL_LINE_LOOP);
    glVertex2f(1,1);
    glVertex2f(1,-1);
    glVertex2f(-1,-1);
    glVertex2f(-1,1);
    glEnd();
    glEnable(GL_DEPTH_TEST);
    resetProjectionMatrix();
}

void MyGLWidget::resizeGL(int width, int height){
    windowWidth = width;
    windowHeight = height;
        qDebug() << "Resize to " << width << "," << height << ", Perspective: " << (double)windowWidth / (double)windowHeight;
    glViewport(0,0,width,height);

    resetProjectionMatrix();

   controller->setScreenScenter(QPoint((double)width/2.0,(double)height/2.0));//lieber zu oft casten bevors nicht tut...
}

void MyGLWidget::resetProjectionMatrix(){

    double p = (double)windowWidth / (double)windowHeight;
    glMatrixMode(GL_PROJECTION);//from HelloGL example
    glLoadIdentity();

    if (isPerspective){
        gluPerspective(45.0,p,0.1,1000);//set vof to 45°
    } else{
        glOrtho(-2*p,2*p,-2,2,1,1000);
    }

    glMatrixMode(GL_MODELVIEW);

}

void MyGLWidget::mousePressEvent(QMouseEvent *event){
    controller->processMousePressEvent(event,this);
}

void MyGLWidget:: mouseMoveEvent(QMouseEvent *event){
    controller->processMouseMoveEvent(event,this);
}

void MyGLWidget::keyPressEvent(QKeyEvent* event){
    if (event->key() == Qt::Key_Control){
        controller->setCtrlPressed(true);
    }
}

void MyGLWidget::keyReleaseEvent(QKeyEvent *event){
    if (event->key() == Qt::Key_Control){
        controller->setCtrlPressed(false);
        qDebug("ctrl released");
    }
}

void MyGLWidget::translateCamera(double xDiff, double yDiff, double zDiff)
{
    //in order to "move parallel to image plane", we have to rotate the image coordinates xDiff and yDiff by our current camera rotation.
    camRotCenter += cameraRotation.conjugate().rotatedVector(QVector3D(xDiff,yDiff,zDiff));
    updateGL();
}

void MyGLWidget::translateQube(double xDiff, double yDiff){
    cubeTranslation += cameraRotation.conjugate().rotatedVector(QVector3D(xDiff, yDiff,0));
    updateGL();
}

void MyGLWidget::rotateQube(QQuaternion diff){

        /*
         * Wikipedia: Two rotation quaternions can be combined into
         * one equivalent quaternionby the relation:
         * q' = q_2*q_1'
         * in which q′ corresponds to the rotation q1 followed by the rotation q2.
         *  (Note that quaternion multiplication is not commutative.)
         */
        qubeRotation = diff*qubeRotation;
        qubeRotation.normalize();

        updateGL();

}

void MyGLWidget::rotateCamera(QQuaternion diff){
    if (isPerspective){
        cameraRotation = diff*cameraRotation;
        cameraRotation.normalize();
        updateGL();
    }
}

//4.1.2 User Input: Zooming
void MyGLWidget::wheelEvent(QWheelEvent *event){
    cameraZoom += 0.01*event->delta();//0.01 seemed to be a good factor
    qDebug() << "Cam zoom delta " << event->delta() << " stored " << cameraZoom;
    updateGL();
}


//4.1.1 slots for shading modes

void MyGLWidget::shadeWireframeSlot(){
    qDebug() << "Wireframe selected";

    phongShader->release();
    glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);//Wireframe looks way cooler with both faces on
    updateGL();
}

void MyGLWidget::shadeFlatSlot(){
    qDebug() << "Flat selected";
    phongShader->release();
    glShadeModel(GL_FLAT);
    glPolygonMode(GL_FRONT,GL_FILL);
    updateGL();
}

void MyGLWidget::shadeGouraudSlot(){
    qDebug() << "Gouraud selected";
    phongShader->release();
    glShadeModel(GL_SMOOTH);
    glPolygonMode(GL_FRONT,GL_FILL);
    updateGL();
}

void MyGLWidget::shadePhongSlot(){
    qDebug() << "Phong selected";
    glPolygonMode(GL_FRONT,GL_FILL);
    phongShader->bind();
    updateGL();
}

//4.1.1 tesselation slot

void MyGLWidget::setTessellation(int t){
    qDebug() << "Tesselation: " << t;
    tesselation = t;
    updateGL();
}

void MyGLWidget::resetCamera(){
    qubeRotation = QQuaternion();
    cubeTranslation = QVector3D(0,0,0);
    camRotCenter = QVector3D(0,0,0);
    cameraZoom = cameraZoomDefault;
    cameraRotation = cameraStartRotation;
    updateGL();
}
